Imports System.io
Public Class CMUCalFunctions

#Region " Global Variables "

    Private UPV As New UPx.Application      'Include UPV driver
    'Private UPVError As Boolean = False
    Private _UPVError As Boolean = False
    Private _UPVErrorMessage As String
    Private Const NaN As Double = 3.402823466385E+38
    Private MeasureBox As New Measuring
    Private _IOCurveXValues() As Double
    Private _IOCurveYValues() As Double

#End Region

#Region " Public Structures "

    Public Structure ValuePair
        Dim AValue As Double
        Dim BValue As Double
    End Structure

#End Region

#Region " UPV Communication "

    Private Function InitiateUpx(ByVal Input As String) As Boolean
        Dim Control As String = String.Empty
        Dim Index As Byte
        Dim Result As Boolean = False
        Dim Counter As Integer = 0
        Dim Instring As String

        Try
            Debug.Write("Trying to initialize UPV remote control...")
            UPV.InitTCP(Input) ' connect UPV
            UPV.Timeout = 60001
            Do                                                                                    'Read error queue until "0; No Error" shows up
                UPV.Write("SYST:ERR?")
                Instring = UPV.Read
                If Instring.StartsWith("0") Then
                    Debug.WriteLine(" succeeded")
                    Result = True
                End If
                Counter += 1
            Loop Until Result Or Counter > 10000

        Catch ex As Exception
            MessageBox.Show("Initialization of UPV remote control failed." & ControlChars.CrLf & "Check whether UPV firmware is running and try again!")
        End Try
        Return Result
    End Function

    Private Sub CloseUpx()
        Debug.Write("Closing UPx server ...")
        UPV.Close() ' disconnect UPV
        Debug.WriteLine(" closed")
        Delay(200)
    End Sub

    Private Sub ClearErrorQueue()
        Dim Instring As String
        Debug.WriteLine("Clearing error queue ...")
        Do
            UPV.Write("SYST:ERR?")
            Instring = UPV.Read
            Debug.WriteLine("Err = " & Instring)
        Loop Until Instring.Substring(0, 1) = "0"
    End Sub

    Private Sub UPVCommand(ByVal CommandString As String)   'Sends command to UPV and checks error queue
        Dim ErrString As String

        UPV.Write(CommandString)
        Debug.Write("UPV Command = " & CommandString)
        UPV.Write("Syst:Err?")
        Do
            ErrString = UPV.Read
            If ErrString = String.Empty Then
                Debug.Write("Empty answer string received: ")
                Debug.WriteLine(UPV.ErrorString)
            End If
        Loop Until ErrString <> String.Empty                'workaround for occasional empty strings
        If Not ErrString.StartsWith("0") Then
            _UPVError = True
        End If
        Debug.WriteLine("   Err = " & ErrString)
    End Sub

    Private Function UPVQuery(ByVal QueryString As String) As String        'Sends a query to UPV, reads response and checks error queue
        Dim AnswerString As String
        Dim ErrString As String

        Try
            UPV.Write(QueryString)
            Debug.Write("UPV Query = " & QueryString)
            Delay(50)
            Do
                AnswerString = UPV.Read
                If AnswerString = String.Empty Then
                    Debug.WriteLine("   Empty answer string received" & "  UPV.ErrorString = " & UPV.ErrorString)
                    UPV.Write(QueryString)
                    Debug.Write("UPV Query = " & QueryString)
                End If
            Loop Until AnswerString <> String.Empty             'workaround for occasional empty strings
            Debug.Write("   Answer = " & AnswerString)
            UPV.Write("Syst:Err?")
            Do
                ErrString = UPV.Read
                If ErrString = String.Empty Then
                    Debug.WriteLine(" Empty answer string received")
                End If
            Loop Until ErrString <> String.Empty
            Debug.WriteLine("   Err = " & ErrString)
            If Not ErrString.StartsWith("0") Then
                _UPVError = True
            End If
            Return AnswerString
        Catch ex As Exception
            MessageBox.Show("UPV query error")
        End Try
    End Function

    Public Sub StartSingleMeasurement()
        InitiateUpx("localhost")
        UPVCommand("Output On")
        UPVCommand("INIT:CONT OFF")
        CloseUpx()
    End Sub

#End Region

#Region " Auxiliary Functions and Subs "

    Private Sub Delay(ByVal DelayTime As Long)
        Dim TimeInstance As Long
        TimeInstance = System.Environment.TickCount
        Do
            Application.DoEvents()
        Loop Until (System.Environment.TickCount - TimeInstance > DelayTime)
    End Sub

    Private Function Average(ByVal Array() As Double) As Double                           'Returns average of all values in the array
        Dim Index As Integer
        Dim Result As Double = 0
        For Index = 0 To Array.Length - 1
            Result = Result + Array(Index)
        Next
        Result = Result / Array.Length
        Return Result
    End Function

    Private Sub Regression(ByVal XArray() As Double, ByVal YArray() As Double, ByRef Slope As Double, ByRef Intercept As Double, ByRef Deviation As Double)      'Calculates straight regression line through points specified by XArray and YArray

        Dim Count As Integer = XArray.Length
        Dim N As Double = 0
        Dim Z As Double = 0
        Dim XAverage As Double = Average(XArray)
        Dim YAverage As Double = Average(YArray)
        Dim S As Double = 0
        Dim Index As Integer

        Try
            For Index = 0 To Count - 1
                Z = Z + (XArray(Index) * YArray(Index))
                N = N + (XArray(Index) * XArray(Index))
            Next
            Z = Z - (Count * XAverage * YAverage)
            N = N - (Count * XAverage * XAverage)
            Slope = Z / N
            Intercept = YAverage - Slope * XAverage
            For Index = 0 To Count - 1
                S = S + Slope * Slope * (XArray(Index) - XAverage) * (XArray(Index) - XAverage) - 2 * Slope * (XArray(Index) - XAverage) * (YArray(Index) - YAverage) + (YArray(Index) - YAverage) * (YArray(Index) - YAverage)
            Next
            Deviation = S / Count
        Catch ex As Exception
            Debug.WriteLine(ex.ToString)
        End Try
    End Sub

    Private Function SubArray(ByVal OriginalArray() As Double, ByVal Length As Integer) As Double()           'Copies the specified number of element to a new array, starting with the first element
        Dim Result() As Double
        Dim Index As Integer

        If (OriginalArray.Length > 0) And (Length <= OriginalArray.Length) Then
            ReDim Result(Length - 1)
            For Index = 0 To Length - 1
                Result(Index) = OriginalArray(Index)
            Next
        End If
        Return Result
    End Function

    Private Function SubArray(ByVal OriginalArray() As Double, ByVal StartIndex As Integer, ByVal Length As Integer) As Double()           'Copies the specified number of element to a new array, starting with the first element
        Dim Result() As Double
        Dim Index As Integer

        If (OriginalArray.Length > 0) And (StartIndex + Length <= OriginalArray.Length) Then
            ReDim Result(Length - 1)
            For Index = 0 To Length - 1
                Result(Index) = OriginalArray(StartIndex + Index)
            Next
        End If
        Return Result
    End Function

#End Region

#Region " Public Functions and Subs "

    Public Function DecoderCal(ByVal SetupFile As String) As Double
        Dim Instring As String
        Dim Time1 As Integer
        Dim Index1 As Integer
        Dim Index2 As Integer
        Dim MeasurementValue(9) As Double
        Dim Max As Double
        Dim Min As Double
        Dim Average As Double
        Dim IsOk As Boolean = False
        Dim Store As Boolean = False
        Dim Answer As DialogResult
        Const Accuracy As Double = 0.2
        Const Repeatability As Double = 0.5
        Dim FSVoltage As Double = NaN         '

        Try
            FSVoltage = NaN
            MeasureBox.InfoText = "decoder output voltage"
            MeasureBox.Show()
            Application.DoEvents()
            InitiateUpx("localhost")
            UPVCommand("*CLS")
            UPVCommand("MMEMory:LOAD:STATe '" & SetupFile & "'")      'Load specific setup
            Debug.WriteLine("Setup file loaded.")
            Time1 = System.Environment.TickCount                                    'Memorize start time for timeout
            Debug.WriteLine("Time1 = " & Time1.ToString)
            For Index1 = 0 To 9                                                     'Evaluate at least 10 values
                Debug.WriteLine("Index1 = " & Index1.ToString)
                UPVCommand("INIT:CONT OFF;*WAI")
                Instring = UPVQuery("SENS:DATA2?")
                Debug.WriteLine("Index1 = " & Index1 & " Value = " & Instring)
                MeasurementValue(Index1) = Val(Instring)
            Next
            Do
                Index1 += 1
                If Index1 > 9 Then Index1 -= 10
                UPVCommand("INIT:CONT OFF;*WAI")
                Instring = UPVQuery("SENS:DATA2?")
                Debug.WriteLine("Index1 = " & Index1 & " Value = " & Instring)
                MeasurementValue(Index1) = Val(Instring)
                Max = -1000
                Min = 1000
                Average = 0
                For Index2 = 0 To 9                                                 'Calculate max, min and average over 10 values
                    If MeasurementValue(Index2) > Max Then Max = MeasurementValue(Index2)
                    If MeasurementValue(Index2) < Min Then Min = MeasurementValue(Index2)
                    Average = Average + MeasurementValue(Index2)
                Next
                Average = Average / 10  'Result in dBV
                IsOk = ((Max - Min) < Accuracy)     'Conditions for valid result
            Loop Until (System.Environment.TickCount > Time1 + 10000) Or IsOk               'Make more measurements if result is not valid
            If IsOk And Not UPVError Then
                FSVoltage = Math.Pow(10, (3 + Average) / 20)
            Else
                FSVoltage = NaN
                MessageBox.Show("Decoder output voltage instable." & ControlChars.CrLf & "Check connections and re-try!", "Decoder Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
            End If
        Catch ex As Exception
            UPVCommand("Output Off")
        End Try
        CloseUpx()
        MeasureBox.Hide()
        Application.DoEvents()
        Return FSVoltage
    End Function

    Public Function EncoderCal(ByVal SetupFile As String, ByVal DecoderCalValue As Double) As Double
        Dim Instring As String
        Dim Time1 As Integer
        Dim Index1 As Integer
        Dim Index2 As Integer
        Dim MeasurementValue(9) As Double
        Dim MobileValue(9) As Double
        Dim Max As Double
        Dim Min As Double
        Dim Average As Double
        Dim IsOk As Boolean = False
        Dim Store As Boolean = False
        Dim Answer As DialogResult
        Const Accuracy As Double = 0.2
        Const Repeatability As Double = 0.5
        Dim InputVoltage As Double = 1000000000.0            '
        Dim EncoderCalValue As Double = NaN
        Dim Result As Double
        Dim GenLevel As Double = -40
        Dim Start As Boolean = True
        Dim Increment As Double

        Try
            EncoderCalValue = NaN
            MeasureBox.InfoText = "encoder input sensitivity"
            MeasureBox.Show()
            Application.DoEvents()
            InitiateUpx("localhost")
            UPVCommand("*CLS")
            UPVCommand("MMEMory:LOAD:STATe '" & SetupFile & "'")      'Load specific setup
            UPVCommand("INPut:CHANnel CH1A")
            Debug.WriteLine("Setup file loaded.")
            UPVCommand("Output On")
            UPVCommand("INIT:CONT OFF;*WAI")
            Instring = UPVQuery("SENS:DATA2?")
            MeasurementValue(1) = Val(Instring)
            Instring = UPVQuery("SENS:DATA?")
            MobileValue(1) = Val(Instring)
            If MeasurementValue(1) < 2 Then
                UPVCommand("SOUR:VOLT " & GenLevel.ToString & " DBV")
                Delay(200)
                UPVCommand("INIT:CONT OFF;*WAI")
                Instring = UPVQuery("SENS:DATA2?")
                If Val(Instring) > -80 Then
                    Do
                        UPVCommand("SOUR:VOLT " & GenLevel.ToString & " DBV")
                        Start = True
                        Do
                            Time1 = System.Environment.TickCount                                    'Memorize start time for timeout
                            Debug.WriteLine("Time1 = " & Time1.ToString)
                            If Start Then
                                For Index1 = 0 To 8                                                     'Evaluate at least 9+1 = 10 values
                                    Debug.WriteLine("Index1 = " & Index1.ToString)
                                    UPVCommand("INIT:CONT OFF;*WAI")
                                    Instring = UPVQuery("SENS:DATA2?")
                                    Debug.WriteLine("Index1 = " & Index1 & " Value = " & Instring)
                                    MeasurementValue(Index1) = Val(Instring)
                                    Instring = UPVQuery("SENS:DATA?")
                                    Debug.WriteLine("Index1 = " & Index1 & " MobileValue = " & Instring)
                                    MobileValue(Index1) = Val(Instring)
                                Next
                                Start = False
                            End If
                            Index1 += 1
                            If Index1 > 9 Then Index1 -= 10
                            UPVCommand("INIT:CONT OFF;*WAI")
                            Instring = UPVQuery("SENS:DATA2?")
                            Debug.WriteLine("Index1 = " & Index1 & " Value = " & Instring)
                            MeasurementValue(Index1) = Val(Instring)
                            Max = -1000
                            Min = 1000
                            Average = 0
                            For Index2 = 0 To 9                                                 'Calculate max, min and average over 10 values
                                If MeasurementValue(Index2) > Max Then Max = MeasurementValue(Index2)
                                If MeasurementValue(Index2) < Min Then Min = MeasurementValue(Index2)
                                Average = Average + MeasurementValue(Index2)
                            Next
                            Average = Average / 10  'Result in dBV
                            Debug.WriteLine("Average = " & Average.ToString)
                            IsOk = ((Max - Min) < Accuracy)     'Conditions for valid result
                            Debug.WriteLine("Time1 = " & Time1.ToString & "  System.Environment.TickCount = " & System.Environment.TickCount.ToString)
                        Loop Until (System.Environment.TickCount > Time1 + 10000) Or IsOk               'Make more measurements if result is not valid
                        Debug.WriteLine("GenLevel = " & GenLevel.ToString)
                        Increment = 20 * Math.Log10(DecoderCalValue) - 13 - Average
                        Debug.WriteLine("Increment = " & Increment.ToString)
                        GenLevel += Increment
                        Debug.WriteLine("GenLevel = " & GenLevel.ToString)
                    Loop Until Math.Abs(Increment) < Accuracy Or Not IsOk
                    If IsOk Then
                        EncoderCalValue = Math.Pow(10, (GenLevel + 13) / 20)                    'V peak
                    Else
                        EncoderCalValue = NaN
                        MessageBox.Show("Encoder loopback voltage instable." & ControlChars.CrLf & "Check connections and re-try!", "Encoder Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                    End If
                Else
                    EncoderCalValue = NaN
                    MessageBox.Show("No signal." & ControlChars.CrLf & "Check connections and re-try!", "Encoder Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                End If
            Else
                EncoderCalValue = NaN
            End If
            UPVCommand("Output Off")
        Catch ex As Exception
            EncoderCalValue = NaN
            UPVCommand("Output Off")
            Debug.WriteLine(ex.ToString)
        End Try
        Result = EncoderCalValue
        CloseUpx()
        MeasureBox.Hide()
        Application.DoEvents()
        Return Result
    End Function

    Public Function MobileOutputCal(ByVal SetupFile As String, ByVal EncoderCalValue As Double) As Double    'Measures peak output from mobile 
        Dim Result As Double = NaN
        Dim Instring As String
        Dim OutputFSLevel As Double

        Try
            MeasureBox.InfoText = "mobile peak output voltage"
            MeasureBox.Show()
            Application.DoEvents()
            InitiateUpx("localhost")
            UPVCommand("*CLS")
            UPVCommand("MMEMory:LOAD:STATe '" & SetupFile & "'")      'Load specific setup
            UPVCommand("INPut:CHANnel CH1")
            Debug.WriteLine("Setup file loaded.")
            UPVCommand("SENSe:FUNCtion PEAK")
            UPVCommand("SENSe:FUNCtion:MMODe PABSolute")
            UPVCommand("SENSe:VOLTage:INTVtime:MODE VALue")
            UPVCommand("SENSe:VOLTage:INTVtime 8.3 S")
            UPVCommand("SOURce:PLAY:SCALepktofs ON")
            OutputFSLevel = 20 * Math.Log10(EncoderCalValue) + 3                                      'Output signal level
            UPVCommand("SOURce:VOLTage:TOTal " & OutputFSLevel.ToString("0.0##") & " DBV")      'set output FS signal level
            UPVCommand("Output On")
            UPVCommand("INIT:CONT RST;*WAI")
            Instring = UPVQuery("SENS:DATA?")
            Result = Val(Instring)

        Catch ex As Exception
            UPVCommand("Output Off")
            Debug.WriteLine(ex.ToString)
        End Try
        CloseUpx()
        MeasureBox.Hide()
        Application.DoEvents()
        Return Result
    End Function

    Public Function MobileInputCal(ByVal SetupFile As String, ByVal DecoderCalValue As Double, ByVal FilterFile As String) As Double
        Dim Instring As String
        Dim Time1 As Integer
        Dim Index1 As Integer
        Dim Index2 As Integer
        Dim MeasurementValue(9) As Double
        Dim MobileValue(9) As Double
        Dim Max As Double
        Dim Min As Double
        Dim MobileMax As Double
        Dim MobileMin As Double
        Dim Average As Double
        Dim IsOk As Boolean = False
        Dim Store As Boolean = False
        Dim Answer As DialogResult
        Dim InputVoltage As Double = 1000000000.0            '
        Dim Result As Double = NaN
        Dim ARL As Double = -60
        Dim Iterations As Byte = 0
        Dim Deviation As Double
        Dim TargetValue As Double = 20 * Math.Log10(DecoderCalValue) - 10 - 3       '-10 dBFS

        Try
            Debug.WriteLine("TargetValue = " & TargetValue.ToString & " dBV")
            Result = NaN
            MeasureBox.InfoText = "mobile input sensitivity"
            MeasureBox.Show()
            Application.DoEvents()
            InitiateUpx("localhost")
            UPVCommand("*CLS")
            UPVCommand("MMEMory:LOAD:STATe '" & SetupFile & "'")      'Load specific setup
            Debug.WriteLine("Setup file loaded.")
            If Not FilterFile = String.Empty Then
                UPVCommand("SENSe:UFILter9 FILE")
                UPVCommand("SENSe:UFILter9:FNAMe '" & FilterFile & "'")
                UPVCommand("SOURce:FILTer UFIL9")
                ARL = -40                                       'Level to speaker higher than to mic input
            End If
            UPVCommand("Output On")

            Do
                UPVCommand("SOURce:VOLTage " & ARL.ToString & " DBV")
                Average = 0
                For Index1 = 1 To 10
                    UPVCommand("Init:Cont OFF;*WAI")
                    Instring = UPVQuery("SENS:DATA2?")
                    Average = Average + Val(Instring)
                Next
                Average = Average / 10
                Deviation = Average - TargetValue
                If (Math.Abs(Deviation)) < 20 Or (Deviation > 0) Then
                    ARL = ARL - Deviation
                Else
                    ARL = ARL + 20                      'increase by 20 dB if step is too large (VAD not activated)
                End If

                Debug.WriteLine("ARL = " & ARL.ToString & " dBV")
                Iterations += 1

            Loop Until (Math.Abs(Deviation) < 1) Or (Iterations = 10) Or (ARL > 20)

            If Math.Abs(Deviation) < 1 Then
                Result = Math.Pow(10, (3 + ARL + 10) / 20)
            Else
                Result = NaN
                MessageBox.Show("Uplink input level not adjustable." & ControlChars.CrLf & "Check connections and re-try!", "Mobile Input Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
            End If
        Catch ex As Exception
            Result = NaN
            Debug.WriteLine(ex.ToString)
        End Try
        UPVCommand("Output Off")
        CloseUpx()
        MeasureBox.Hide()
        Application.DoEvents()
        Return Result

    End Function

    Public Function MobileInputSpeechCal(ByVal SetupFile As String, ByVal DecoderCalValue As Double, ByVal FilterFile As String, ByVal ReducedLevel As Boolean) As Double
        Dim Result As Double = NaN
        Dim Instring As String
        Dim GenVolt As Double = 0.01
        Dim RMS As Double
        Dim Iterations As Byte = 0
        '               peak decoder output   16 dBm0 from peak                activity factor
        Dim TargetRMS As Double = DecoderCalValue / (Math.Pow(10, 19.14 / 20) * Math.Pow(10, 3.83 / 20))
        Dim TargetLevel As Double
        Dim MeasuredLevel As Double
        Dim OldMeasuredLevel As Double
        Dim UpperGenLevel As Double
        Dim LowerGenLevel As Double
        Dim UpperMeasuredLevel As Double
        Dim LowerMeasuredLevel As Double
        Dim OldGenLevel As Double = 20 * Math.Log10(GenVolt)
        Dim NewGenLevel As Double

        Try
            If ReducedLevel Then TargetRMS = DecoderCalValue / (Math.Pow(10, 29.14 / 20) * Math.Pow(10, 3.83 / 20))
            Debug.WriteLine("TargetRMS = " & TargetRMS.ToString)
            TargetLevel = 20 * Math.Log10(TargetRMS)
            Debug.WriteLine("TargetLevel = " & TargetLevel.ToString)
            MeasureBox.InfoText = "signal rms"
            MeasureBox.Show()
            Application.DoEvents()
            InitiateUpx("localhost")
            UPVCommand("*CLS")
            UPVCommand("MMEMory:LOAD:STATe '" & SetupFile & "'")      'Load specific setup
            Debug.WriteLine("Setup file loaded.")
            If Not FilterFile = String.Empty Then
                UPVCommand("SENSe:UFILter9 FILE")
                UPVCommand("SENSe:UFILter9:FNAMe '" & FilterFile & "'")
                UPVCommand("SOURce:FILTer UFIL9")
                GenVolt = 0.1                                       'Level to speaker higher than to mic input
            End If

            '***** Set RMS measurement *****

            UPVCommand("SENSe:VOLTage:RANGe2:VALue " & DecoderCalValue)
            UPVCommand("SENSe:FUNCtion RMS")
            UPVCommand("SENSe:FUNCtion:APERture:MODE VAL")
            UPVCommand("SENSe:VOLTage:APERture 8.30850028992 S")

            '***** Set Generator *****

            UPVCommand("SOURce:PLAY:MODE TOS")
            UPVCommand("SOURce:PLAY:RESTart OFF")
            Debug.WriteLine("GenVolt = " & GenVolt.ToString)
            UPVCommand("SOURce:VOLTage " & GenVolt.ToString & " V")
            UPVCommand("Output On")

            '***** Start first measurement *****

            UPV.Write("INIT:CONT RST;*WAI")                     'start measurement and restart generator
            Instring = UPVQuery("Sens:Data2?")
            RMS = Val(Instring)
            Debug.WriteLine("RMS = " & RMS.ToString)
            MeasuredLevel = 20 * Math.Log10(RMS)
            Debug.WriteLine("MeasuredLevel = " & MeasuredLevel.ToString)

            '***** Check level difference ****

            Do While Math.Abs(TargetLevel - MeasuredLevel) > 10
                OldMeasuredLevel = MeasuredLevel
                GenVolt = 10 * GenVolt
                NewGenLevel = 20 * Math.Log10(GenVolt)
                UPVCommand("SOURce:VOLTage " & GenVolt.ToString & " V")
                UPV.Write("INIT:CONT RST;*WAI")                     'start measurement and restart generator
                Instring = UPVQuery("Sens:Data2?")
                RMS = Val(Instring)
                Debug.WriteLine("RMS = " & RMS.ToString)
                MeasuredLevel = 20 * Math.Log10(RMS)
                Debug.WriteLine("MeasuredLevel = " & MeasuredLevel.ToString)
                If Math.Abs(MeasuredLevel - OldMeasuredLevel) < 2 Then
                    Result = NaN
                    UPVCommand("SOURce:PLAY:RESTart Auto")
                    UPVCommand("Output Off")
                    CloseUpx()
                    MeasureBox.Hide()
                    Application.DoEvents()
                    MessageBox.Show("No reaction to change of input voltage." & ControlChars.CrLf & "Check connections and re-try!", "Mobile Input Speech Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                    Return Result
                End If
                OldGenLevel = NewGenLevel
            Loop

            UpperGenLevel = OldGenLevel + TargetLevel - MeasuredLevel + 5
            Debug.WriteLine("UpperGenLevel = " & UpperGenLevel.ToString)
            LowerGenLevel = OldGenLevel + TargetLevel - MeasuredLevel - 5
            Debug.WriteLine("LowerGenLevel = " & LowerGenLevel.ToString)

            GenVolt = Math.Pow(10, UpperGenLevel / 20)
            If GenVolt < 10 Then
                UPVCommand("SOURce:VOLTage " & GenVolt.ToString & " V")
                UPV.Write("INIT:CONT RST;*WAI")                     'start measurement and restart generator
                Instring = UPVQuery("Sens:Data2?")
                RMS = Val(Instring)
                Debug.WriteLine("RMS = " & RMS.ToString)
            End If
            UpperMeasuredLevel = 20 * Math.Log10(RMS)
            Debug.WriteLine("UpperMeasuredLevel = " & UpperMeasuredLevel.ToString)

            GenVolt = Math.Pow(10, LowerGenLevel / 20)
            If GenVolt < 10 Then
                UPVCommand("SOURce:VOLTage " & GenVolt.ToString & " V")
                UPV.Write("INIT:CONT RST;*WAI")                     'start measurement and restart generator
                Instring = UPVQuery("Sens:Data2?")
                RMS = Val(Instring)
                Debug.WriteLine("RMS = " & RMS.ToString)
            End If
            LowerMeasuredLevel = 20 * Math.Log10(RMS)
            Debug.WriteLine("LowerMeasuredLevel = " & LowerMeasuredLevel.ToString)

            NewGenLevel = OldGenLevel + TargetLevel - MeasuredLevel
            Debug.WriteLine("NewGenLevel = " & NewGenLevel.ToString)
            OldGenLevel = NewGenLevel

            Do
                Iterations += 1
                GenVolt = Math.Pow(10, NewGenLevel / 20)
                Debug.WriteLine("GenVolt = " & GenVolt.ToString)
                If GenVolt < 10 Then
                    UPVCommand("SOURce:VOLTage " & GenVolt.ToString & " V")
                    UPV.Write("INIT:CONT RST;*WAI")                     'start measurement and restart generator
                    Instring = UPVQuery("Sens:Data2?")
                    RMS = Val(Instring)
                    Debug.WriteLine("RMS = " & RMS.ToString)
                End If
                MeasuredLevel = 20 * Math.Log10(RMS)
                Debug.WriteLine("MeasuredLevel = " & MeasuredLevel.ToString)
                If MeasuredLevel > TargetLevel Then
                    NewGenLevel = OldGenLevel - 5 * (MeasuredLevel - TargetLevel) / (UpperMeasuredLevel - TargetLevel)
                Else
                    NewGenLevel = OldGenLevel + 5 * (TargetLevel - MeasuredLevel) / (TargetLevel - LowerMeasuredLevel)
                End If
                Debug.WriteLine("NewGenLevel = " & NewGenLevel.ToString)
                OldGenLevel = NewGenLevel

            Loop Until ((RMS / TargetRMS > 0.9) And (RMS / TargetRMS < 1.1)) Or (Iterations > 5) Or (GenVolt >= 10)

            If (Iterations > 5) Then
                MeasureBox.Hide()
                MessageBox.Show("Mobile input voltage not adjustable." & ControlChars.CrLf & "Check connections, disable AGC and re-try!", "Mobile Input Speech Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
            Else
                If (GenVolt >= 10) Then
                    MeasureBox.Hide()
                    MessageBox.Show("Mobile input sensitivity too low." & ControlChars.CrLf & "Check connections, and re-try!", "Mobile Input Speech Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                Else
                    '       peak gen. output  peak from -16 dBm0   activity factor        peak from rms
                    Result = GenVolt * Math.Pow(10, 19.14 / 20) * Math.Pow(10, 3.83 / 20) * 0.0301 'FS input peak of the mobile
                    If ReducedLevel Then Result = Result * Math.Pow(10, 10 / 20)
                End If
            End If

        Catch ex As Exception
            Result = NaN
            Debug.WriteLine(ex.ToString)
        End Try

        UPVCommand("SOURce:PLAY:RESTart Auto")
        UPVCommand("Output Off")
        CloseUpx()
        MeasureBox.Hide()
        Application.DoEvents()
        Return Result

    End Function

    Public Function InOutCal(ByVal SetupFile As String, ByVal Output As Byte, ByVal Input As Byte, ByVal GenFilterFile As String, ByVal AnFilterFile As String) As ValuePair
        Dim Result As ValuePair
        Dim InValue(40) As Double
        Dim OutValue(40) As Double
        Dim Index As Byte
        Dim Instring As String
        Dim XSubarray() As Double
        Dim YSubarray() As Double
        Dim Slope1(40) As Double
        Dim Intercept1(40) As Double
        Dim Residue1(40) As Double
        Dim Slope2(40) As Double
        Dim Intercept2(40) As Double
        Dim Residue2(40) As Double
        Dim MinRegressionIndex1 As Byte
        Dim MinRegressionIndex2 As Byte
        Dim MinRegression As Double = 1000
        Dim IsOk As Boolean = False

        Try
            InitiateUpx("localhost")
            MeasureBox.Show()
            Application.DoEvents()
            UPVCommand("*CLS")
            UPVCommand("MMEMory:LOAD:STATe '" & SetupFile & "'")      'Load specific setup            Debug.WriteLine("Setup file loaded.")
            UPVCommand("OUTPut:CHANnel CH" & Output.ToString)
            UPVCommand("INPut:CHANnel CH" & Input.ToString)
            If Not GenFilterFile = String.Empty Then
                UPVCommand("SENSe:UFILter9 FILE")
                UPVCommand("SENSe:UFILter9:FNAMe '" & GenFilterFile & "'")
                UPVCommand("SOURce:FILTer UFIL9")
            End If
            If Not AnFilterFile = String.Empty Then
                UPVCommand("SENSe:UFILter8 FILE")
                UPVCommand("SENSe:UFILter8:FNAMe '" & AnFilterFile & "'")
                UPVCommand("INPut:FILTer UFIL8")
            End If
            UPVCommand("Output On")

            For Index = 0 To 40
                InValue(40 - Index) = 10 - 2 * Index
                MeasureBox.InfoText = "output level at " & InValue(40 - Index).ToString & " dBV"
                Application.DoEvents()
                UPVCommand("SOURce:VOLTage " & InValue(40 - Index).ToString & " DBV")
                UPVCommand("INIT:CONT OFF;*WAI")
                Instring = UPVQuery("SENS:DATA" & Input.ToString & "?")
                OutValue(40 - Index) = Val(Instring)
            Next
            _IOCurveXValues = InValue
            _IOCurveYValues = OutValue

            MeasureBox.InfoText = "calculating ascending line"
            Application.DoEvents()
            For Index = 0 To 20                             'calculate regression lines for 10 values each
                XSubarray = SubArray(InValue, Index, 10)
                YSubarray = SubArray(OutValue, Index, 10)
                Regression(XSubarray, YSubarray, Slope1(Index), Intercept1(Index), Residue1(Index))           'Calculate straight regression line
                Debug.WriteLine("Index = " & Index.ToString & "  Slope = " & Slope1(Index).ToString & "  Residue = " & Residue1(Index).ToString)
                If Residue1(Index) < MinRegression Then      'find best fitting line over 10 values
                    MinRegression = Residue1(Index)
                    MinRegressionIndex1 = Index
                End If
            Next

            MeasureBox.InfoText = "calculating saturation line"
            Application.DoEvents()
            MinRegression = 100
            For Index = 4 To 20                             'calculate regression lines for the highest generator values
                XSubarray = SubArray(InValue, 40 - Index - 1, Index)
                YSubarray = SubArray(OutValue, 40 - Index - 1, Index)
                Regression(XSubarray, YSubarray, Slope2(Index), Intercept2(Index), Residue2(Index))           'Calculate straight regression line
                Debug.WriteLine("Index = " & Index.ToString & "  Slope = " & Slope2(Index).ToString & "  Residue = " & Residue2(Index).ToString)
                If Residue2(Index) < MinRegression Then      'find best fitting line
                    MinRegression = Residue2(Index)
                    MinRegressionIndex2 = Index
                End If
            Next

            IsOk = ((Slope1(MinRegressionIndex1) > 0.8) And (Slope1(MinRegressionIndex1) < 1.2))

            If IsOk Then
                Result.AValue = Math.Pow(10, ((3 + Intercept2(MinRegressionIndex2) - Intercept1(MinRegressionIndex1)) / Slope1(MinRegressionIndex1)) / 20)       'FS input peak value
                Result.BValue = Math.Pow(10, ((3 + Intercept2(MinRegressionIndex2)) / 20))                                                                     'FS output peak value
            Else
                MessageBox.Show("No signal." & ControlChars.CrLf & "Check connections and re-try!", "In/Out Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                Result.AValue = NaN                         'invalid result
                Result.BValue = NaN                         'invalid result
            End If

            Debug.WriteLine("Result.AValue = " & Result.AValue.ToString)
            Debug.WriteLine("Result.BValue = " & Result.BValue.ToString)

        Catch ex As Exception
            Result.AValue = NaN
            Result.BValue = NaN
            UPVCommand("Output Off")
            Debug.WriteLine(ex.ToString)
        End Try
        UPVCommand("Output Off")
        MeasureBox.Hide()
        Application.DoEvents()
        Return Result
    End Function

    Public Function InOutSpeechCal(ByVal SetupFile As String, ByVal Output As Byte, ByVal Input As Byte, ByVal GenFilterFile As String, ByVal AnFilterFile As String) As ValuePair
        Dim Result As ValuePair
        Dim InValue(30) As Double
        Dim OutValue(30) As Double
        Dim Index As Byte
        Dim Instring As String
        Dim XSubarray() As Double
        Dim YSubarray() As Double
        Dim Slope1(30) As Double
        Dim Intercept1(30) As Double
        Dim Residue1(30) As Double
        Dim Slope2(30) As Double
        Dim Intercept2(30) As Double
        Dim Residue2(30) As Double
        Dim MinRegressionIndex1 As Byte
        Dim MinRegressionIndex2 As Byte
        Dim MinRegression As Double = 1000
        Dim IsOk As Boolean = False

        Try
            InitiateUpx("localhost")
            MeasureBox.Show()
            Application.DoEvents()
            UPVCommand("*CLS")

            '***** Load specific setup *****

            UPVCommand("MMEMory:LOAD:STATe '" & SetupFile & "'")
            Debug.WriteLine("Setup file loaded.")
            UPVCommand("OUTPut:CHANnel CH" & Output.ToString)
            UPVCommand("INPut:CHANnel CH" & Input.ToString)

            '***** Set RMS measurement *****

            UPVCommand("SENSe:VOLTage:RANGe2:VALue 3V")
            UPVCommand("SENSe:FUNCtion RMS")
            UPVCommand("SENSe:FUNCtion:APERture:MODE VAL")
            UPVCommand("SENSe:VOLTage:APERture 8.30850028992 S")
            UPVCommand("SENSe:UNIT" & Input.ToString & " DBV")

            '***** Set Filters *****

            If Not GenFilterFile = String.Empty Then
                UPVCommand("SENSe:UFILter9 FILE")
                UPVCommand("SENSe:UFILter9:FNAMe '" & GenFilterFile & "'")
                UPVCommand("SOURce:FILTer UFIL9")
            End If
            If Not AnFilterFile = String.Empty Then
                UPVCommand("SENSe:UFILter8 FILE")
                UPVCommand("SENSe:UFILter8:FNAMe '" & AnFilterFile & "'")
                UPVCommand("INPut:FILTer UFIL8")
            End If

            '***** Set Generator *****

            UPVCommand("SOURce:FUNCtion PLAY")
            UPVCommand("OUTPut:BANDwidth:MODE AUTO")
            UPVCommand("SOURce:PLAY:SCALepktofs ON")
            UPVCommand("MMEMory:LOAD:ARBitrary 'C:/UPV/Config/ref/PesqRef.wav'")
            UPVCommand("SOURce:PLAY:MODE TOS")
            UPVCommand("SOURce:PLAY:RESTart OFF")
            UPVCommand("Output On")

            For Index = 0 To 30
                InValue(30 - Index) = 10 - 2 * Index                    '-50 dBV to +10 dBV
                MeasureBox.InfoText = "output level at " & InValue(30 - Index).ToString & " dBV"
                Application.DoEvents()
                UPVCommand("SOURce:VOLTage " & InValue(30 - Index).ToString & " DBV")
                UPV.Write("INIT:CONT RST;*WAI")                     'start measurement and restart generator
                Instring = UPVQuery("Sens:Data2?")
                OutValue(30 - Index) = Val(Instring) + 24.7         'Add crest factor to calculate peak from rms
            Next
            _IOCurveXValues = InValue
            _IOCurveYValues = OutValue

            MeasureBox.InfoText = "calculating ascending line"
            Application.DoEvents()
            For Index = 0 To 15                             'calculate regression lines for 10 values each
                XSubarray = SubArray(InValue, Index, 10)
                YSubarray = SubArray(OutValue, Index, 10)
                Regression(XSubarray, YSubarray, Slope1(Index), Intercept1(Index), Residue1(Index))           'Calculate straight regression line
                Debug.WriteLine("Index = " & Index.ToString & "  Slope = " & Slope1(Index).ToString & "  Residue = " & Residue1(Index).ToString)
                If Residue1(Index) < MinRegression Then      'find best fitting line over 10 values
                    MinRegression = Residue1(Index)
                    MinRegressionIndex1 = Index
                End If
            Next

            MeasureBox.InfoText = "calculating saturation line"
            Application.DoEvents()
            MinRegression = 100
            For Index = 4 To 20                             'calculate regression lines for the highest generator values
                XSubarray = SubArray(InValue, 30 - Index - 1, Index)
                YSubarray = SubArray(OutValue, 30 - Index - 1, Index)
                Regression(XSubarray, YSubarray, Slope2(Index), Intercept2(Index), Residue2(Index))           'Calculate straight regression line
                Debug.WriteLine("Index = " & Index.ToString & "  Slope = " & Slope2(Index).ToString & "  Residue = " & Residue2(Index).ToString)
                If Residue2(Index) < MinRegression Then      'find best fitting line
                    MinRegression = Residue2(Index)
                    MinRegressionIndex2 = Index
                End If
            Next

            IsOk = ((Slope1(MinRegressionIndex1) > 0.8) And (Slope1(MinRegressionIndex1) < 1.2))

            If IsOk Then
                Result.AValue = Math.Pow(10, ((3 + Intercept2(MinRegressionIndex2) - Intercept1(MinRegressionIndex1)) / Slope1(MinRegressionIndex1)) / 20)       'FS input peak value
                Result.BValue = Math.Pow(10, ((3 + Intercept2(MinRegressionIndex2)) / 20))                                                                     'FS output peak value
            Else
                MessageBox.Show("No signal." & ControlChars.CrLf & "Check connections and re-try!", "In/Out Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                Result.AValue = NaN                         'invalid result
                Result.BValue = NaN                         'invalid result
            End If

            Debug.WriteLine("Result.AValue = " & Result.AValue.ToString)
            Debug.WriteLine("Result.BValue = " & Result.BValue.ToString)

        Catch ex As Exception
            Result.AValue = NaN
            Result.BValue = NaN
            UPVCommand("Output Off")
            Debug.WriteLine(ex.ToString)
        End Try
        UPVCommand("Output Off")
        MeasureBox.Hide()
        Application.DoEvents()
        Return Result
    End Function

    Public Function InOutSpeechCal2(ByVal SetupFile As String, ByVal Output As Byte, ByVal Input As Byte) As ValuePair
        Dim Result As ValuePair
        Dim InValue(40) As Double
        Dim OutValue(40) As Double
        Dim Index As Byte
        Dim Instring As String
        Dim XSubarray() As Double
        Dim YSubarray() As Double
        Dim Slope1(40) As Double
        Dim Intercept1(40) As Double
        Dim Residue1(40) As Double
        Dim Slope2(40) As Double
        Dim Intercept2(40) As Double
        Dim Residue2(40) As Double
        Dim MinRegressionIndex1 As Byte
        Dim MinRegressionIndex2 As Byte
        Dim MinRegression As Double = 1000
        Dim IsOk As Boolean = False
        Dim XInterpol As Double
        Dim YInterpol As Double
        Dim Slope As Double

        Try
            InitiateUpx("localhost")
            MeasureBox.Show()
            Application.DoEvents()
            UPVCommand("*CLS")
            UPVCommand("MMEMory:LOAD:STATe '" & SetupFile & "'")      'Load specific setup            Debug.WriteLine("Setup file loaded.")
            UPVCommand("OUTPut:CHANnel CH" & Output.ToString)
            UPVCommand("INPut:CHANnel CH" & Input.ToString)
            UPVCommand("Output On")

            '***** Set RMS measurement *****

            UPVCommand("SENSe:VOLTage:RANGe2:VALue 3V")
            UPVCommand("SENSe:FUNCtion RMS")
            UPVCommand("SENSe:FUNCtion:APERture:MODE VAL")
            UPVCommand("SENSe:VOLTage:APERture 8.30850028992 S")
            UPVCommand("SENSe:UNIT" & Input.ToString & " DBV")

            '***** Set Generator *****

            UPVCommand("SOURce:PLAY:MODE TOS")
            UPVCommand("SOURce:PLAY:RESTart OFF")
            UPVCommand("Output On")

            For Index = 0 To 40
                InValue(40 - Index) = 20 - 2 * Index
                MeasureBox.InfoText = "output level at " & InValue(40 - Index).ToString & " dBV"
                Application.DoEvents()
                UPVCommand("SOURce:VOLTage " & InValue(40 - Index).ToString & " DBV")
                UPV.Write("INIT:CONT RST;*WAI")                     'start measurement and restart generator
                Instring = UPVQuery("SENS:DATA" & Input.ToString & "?")
                OutValue(40 - Index) = Val(Instring)
            Next

            _IOCurveXValues = InValue
            _IOCurveYValues = OutValue

            MeasureBox.InfoText = "calculating ascending line"
            Application.DoEvents()
            For Index = 0 To 20                             'calculate regression lines for 10 values each
                XSubarray = SubArray(InValue, Index, 10)
                YSubarray = SubArray(OutValue, Index, 10)
                Regression(XSubarray, YSubarray, Slope1(Index), Intercept1(Index), Residue1(Index))           'Calculate straight regression line
                Debug.WriteLine("Index = " & Index.ToString & "  Slope = " & Slope1(Index).ToString & "  Residue = " & Residue1(Index).ToString)
                If Residue1(Index) < MinRegression Then      'find best fitting line over 10 values
                    MinRegression = Residue1(Index)
                    MinRegressionIndex1 = Index
                End If
            Next

            MeasureBox.InfoText = "Calculating 1 dB compression point"
            Application.DoEvents()

            Index = 10
            Do
                Index += 1
            Loop Until (OutValue(Index) < Intercept1(MinRegressionIndex1) + Slope1(MinRegressionIndex1) * InValue(Index) - 1) Or Index > 40

            IsOk = ((Slope1(MinRegressionIndex1) > 0.8) And (Slope1(MinRegressionIndex1) < 1.2)) And (Index < 41)

            If IsOk Then

                '***** Interpolation *****

                Slope = (OutValue(Index) - OutValue(Index - 1)) / (InValue(Index) - InValue(Index - 1))
                XInterpol = (OutValue(Index) - Slope * InValue(Index) - Intercept1(MinRegressionIndex1) + 1) / (Slope1(MinRegressionIndex1) - Slope)
                YInterpol = Intercept1(MinRegressionIndex1) - 1 + Slope1(MinRegressionIndex1) * XInterpol
                Result.AValue = Math.Pow(10, (XInterpol - 12) / 20)   'FS input peak value
                Result.BValue = Math.Pow(10, (YInterpol + 20) / 20)   'FS decoder output peak value                                                                'FS output peak value
            Else
                MessageBox.Show("No signal." & ControlChars.CrLf & "Check connections and re-try!", "In/Out Calibration", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                Result.AValue = NaN                         'invalid result
                Result.BValue = NaN                         'invalid result
            End If

            Debug.WriteLine("Result.AValue = " & Result.AValue.ToString)
            Debug.WriteLine("Result.BValue = " & Result.BValue.ToString)

        Catch ex As Exception
            Result.AValue = NaN
            Result.BValue = NaN
            UPVCommand("Output Off")
            Debug.WriteLine(ex.ToString)
        End Try
        UPVCommand("Output Off")
        UPVCommand("SOURce:PLAY:RESTart Auto")
        MeasureBox.Hide()
        Application.DoEvents()
        Return Result
    End Function

    Public Sub LoadSetup(ByVal Output As Byte, ByVal Input As Byte, ByVal OutCal As Double, ByVal InCal As Double, ByVal SetupFile As String, ByVal SignalFile As String, ByVal Gain As Double, ByVal GenFiltFile As String, ByVal AnFiltFile As String, ByVal ScaleToFS As Boolean)
        Dim OutputFSLevel As Double
        Try
            If File.Exists(SetupFile) Then
                InitiateUpx("localhost")
                UPVCommand("MMEMory:LOAD:STATe '" & SetupFile & "'")                                'load setup
                If SignalFile <> "" Then
                    If File.Exists(SignalFile) Then
                        UPVCommand("MMEMory:LOAD:ARBitrary '" & SignalFile & "'")                           'load signal file
                    Else
                        MessageBox.Show("Specified wave file does not exist." & ControlChars.CrLf & "Open existing wave file and re-try!", "Load Setup", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
                    End If
                End If
                UPVCommand("OUTPut:CHANnel CH" & Output.ToString)                                   'set output channel
                UPVCommand("INPut:CHANnel CH" & Input.ToString)                                     'set input channel
                OutputFSLevel = 20 * Math.Log10(OutCal) + Gain                                      'Output signal level
                UPVCommand("SOURce:VOLTage:TOTal " & OutputFSLevel.ToString("0.0##") & " DBV")      'set output FS signal level
                UPVCommand("SENSe:VOLTage:RANGe" & Input.ToString & ":VALue " & (0.71 * InCal).ToString)
                If Not GenFiltFile = String.Empty Then
                    UPVCommand("SENSe:UFILter9 FILE")
                    UPVCommand("SENSe:UFILter9:FNAMe '" & GenFiltFile & "'")
                    UPVCommand("SOURce:FILTer UFIL9")
                End If
                If Not AnFiltFile = String.Empty Then
                    UPVCommand("SENSe:UFILter8 FILE")
                    UPVCommand("SENSe:UFILter8:FNAMe '" & AnFiltFile & "'")
                    UPVCommand("INPut:FILTer UFIL8")
                End If
                If ScaleToFS Then
                    UPVCommand("SOURce:PLAY:SCALepktofs ON")
                Else
                    UPVCommand("SOURce:PLAY:SCALepktofs OFF")
                End If
            Else
                MessageBox.Show("Specified setup file does not exist." & ControlChars.CrLf & "Open existing setup file and re-try!", "Load Setup", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
            End If
        Catch ex As Exception
            Debug.WriteLine(ex.ToString)
            MessageBox.Show("Error while loading setup.", "Load Setup", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End Try
        CloseUpx()
    End Sub

    Public Sub ResetUPVErrorInfo()
        _UPVError = False
        _UPVErrorMessage = String.Empty
    End Sub

#End Region

#Region " Public Properties "

    Public ReadOnly Property IOCurveXValues() As Double()
        Get
            Return _IOCurveXValues
        End Get
    End Property

    Public ReadOnly Property IOCurveYValues() As Double()
        Get
            Return _IOCurveYValues
        End Get
    End Property

    Public ReadOnly Property UPVError() As Boolean
        Get
            Return _UPVError
        End Get
    End Property

    Public ReadOnly Property UPVErrorMessage() As String
        Get
            Return _UPVErrorMessage
        End Get
    End Property

#End Region

End Class
